"""

This function can be used to assess the accuracy of emulation of a test data
set given a trained model and produces a figure showing the
mean, 95th percentile and worst emulations. Examples of these figures can be
found in the `MNRAS preprint <https://arxiv.org/abs/2104.04336)>`__. The
figure will be saved in the provided ``'base_dir/'``.

"""

import numpy as np
from globalemu.losses import loss_functions
import matplotlib.pyplot as plt


class signal_plot():

    r"""

    The class can be initialised with the following kwargs and the
    following code

    .. code:: python

        plotter  = signal_plot(parameters, labels, loss_type,
                        predictor, base_dir, **kwargs)

    **Parameters:**

        parameters: **list or np.array**
            | The astrophysical parameters corresponding to the testing data.

        labels: **list or np.array**
            | The signals, corresponding to the input parameters, that
                we want to predict and subsequently plot the mean, 95th
                percentile and worst emulations of.

        loss_type: ** str or function**
            | The metric by which we want to assess the accuracy of emulation.
                The built in loss functions can be accessed by setting this
                variable to 'rmse', 'mse' or 'GEMLoss'. Alternatively, a user
                defined callable function that takes in the labels and signals
                can also be provided.

        predictor: ** globalemu.eval object **
            | An instance of the globalemu eval class that will be used to make
                predictions of the labels from the input parameters.

        base_dir: **string / default: 'model_dir/'**
            | The ``base_dir`` is where the signal plot will be saved.

    **kwargs:**

        rtol: **int or float / default: 1e-2**
            | The relative accuracy with which the function finds a
                signal with a loss equal to the mean loss for all predictions.

        atol: **int or float / default: 1e-2**
            | The absolute accuracy with which the function finds a
                signal with a loss equal to the mean loss for all predictions.

        figsizex: **int or float / default: 5**
            | The of the figure along the x axis to be passed to
                plt.subplots().

        figsizey: **int or float / default: 10**
            | The of the figure along the y axis to be passed to
                plt.subplots().

        xHI: **Bool / default: False**
            | If True then ``globalemu`` will act as if it is evaluating a
                neutral fraction history emulator.

        loss_label: **string/ default: 'Loss = {:.3f}'**
            | This kwarg can be used to adjust the loss labels in the plot
                legends. For example if we wanted precision in the 4th
                decimal place we can set ``loss_label= 'Loss = {:.4f}'``.
                Equally if we wanted to change the name of the loss and add in
                units we can have ``loss_label= 'RMSE = {:.3f} mK'``.

    """

    def __init__(self, parameters, labels, loss_type,
                 predictor, base_dir, **kwargs):

        for key, values in kwargs.items():
            if key not in set(
                    ['xHI', 'rtol',
                     'atol', 'figsizex', 'figsizey', 'loss_label']):
                raise KeyError("Unexpected keyword argument in evaluate()")

        self.rtol = kwargs.pop('rtol', 1e-2)
        self.atol = kwargs.pop('atol', 1e-2)
        self.figsizex = kwargs.pop('figsizex', 5)
        self.figsizey = kwargs.pop('figsizey', 10)
        self.loss_label = kwargs.pop('loss_label', 'Loss = {:.3f}')

        if type(self.loss_label) is not str:
            raise TypeError("'loss_label' must be a string.")

        float_kwargs = [self.rtol, self.atol, self.figsizex, self.figsizey]
        float_kwarg_str = ['rtol', 'atol', 'figsizex', 'figsizey']
        for i in range(len(float_kwargs)):
            if type(float_kwargs[i]) not in set([float, int]):
                raise TypeError("'" + float_kwarg_str[i] +
                                "' must be an integer or a float.")

        self.parameters = parameters

        if type(self.parameters) not in set([np.ndarray, list]):
            raise TypeError("'parameters' must be a list or np.array.")

        self.labels = labels

        if type(self.labels) not in set([np.ndarray, list]):
            raise TypeError("'labels' must be a list or np.array.")

        self.loss_type = loss_type
        self.base_dir = base_dir

        if self.loss_type not in set(['rmse', 'mse', 'GEMLoss']):
            if not callable(self.loss_type):
                raise TypeError("'loss_type' must be a string from the " +
                                "predefined set (see documentaiton) or a " +
                                "user defined function.")

        if type(self.base_dir) is not str:
            raise TypeError("'base_dir' must be a string.")
        elif self.base_dir.endswith('/') is False:
            raise TypeError("'base_dir' must end with '/'.")

        self.predictor = predictor

        if not callable(predictor):
            raise TypeError("'predictor' should be an instance of " +
                            "globalemu.eval.")

        self.xHI = kwargs.pop('xHI', False)

        if type(self.xHI) is not bool:
            raise TypeError("'xHI' should be either True or False.")

        signal, z = self.predictor(self.parameters)

        loss = []
        for i in range(len(signal)):
            if type(self.loss_type) is not str:
                loss.append(self.loss_type(self.labels[i], signal[i]))
            else:
                lf = loss_functions(self.labels[i], signal[i])
                if self.loss_type == 'rmse':
                    loss.append(lf.rmse().numpy())
                elif self.loss_type == 'mse':
                    loss.append(lf.mse())
                elif self.loss_type == 'GEMLoss':
                    loss.append(lf.GEMLoss())
        loss = np.array(loss)

        mean_label = self.labels[
                    np.where(
                             np.isclose(
                                        loss, loss.mean(),
                                        rtol=self.rtol,
                                        atol=self.atol))[0][0], :]
        mean_pred = signal[
                    np.where(
                             np.isclose(
                                        loss, loss.mean(),
                                        rtol=self.rtol,
                                        atol=self.atol))[0][0], :]
        worst_label = self.labels[np.where(loss == loss.max())[0][0], :]
        worst_pred = signal[np.where(loss == loss.max())[0][0], :]

        args = np.argsort(loss)
        sorted_loss = loss[args]
        sorted_labels = self.labels[args]
        sorted_signals = signal[args]
        idx = int(len(sorted_loss)/100*95)
        limit95 = sorted_loss[idx]
        limit_label = sorted_labels[idx, :]
        limit_pred = sorted_signals[idx, :]

        fig, axes = plt.subplots(3, 1,
                                 figsize=(self.figsizex, self.figsizey),
                                 sharex=True)
        axes[0].plot(z, mean_label, label='True Signal')
        axes[0].plot(z, mean_pred, label=self.loss_label.format(
            loss[
                 np.where(np.isclose(loss, loss.mean(),
                          rtol=self.rtol, atol=self.atol))[0][0]]))
        axes[1].plot(z, limit_label, label='True Signal')
        axes[1].plot(z, limit_pred, label=self.loss_label.format(limit95))
        axes[2].plot(z, worst_label, label='True Signal')
        axes[2].plot(z, worst_pred, label=self.loss_label.format(loss.max()))
        axes[0].legend(title='Mean:')
        axes[1].legend(title='95%:')
        axes[2].legend(title='Worst:')

        if self.xHI is False:
            for i in range(len(axes)):
                axes[i].set_ylabel(r'$T_{21}$ [mK]')
        else:
            for i in range(len(axes)):
                axes[i].set_ylabel(r'$x_{HI}$')
        fig.add_subplot(111, frame_on=False)
        plt.tick_params(bottom=False, left=False, labelcolor='none')
        plt.xlabel('$z$')
        plt.tight_layout()
        plt.subplots_adjust(hspace=0, wspace=0)
        plt.savefig(self.base_dir + 'eval_plot.pdf')
        plt.close()



import numpy as np
from globalemu.gui_config import config
from globalemu.downloads import download
import shutil
import os
import pytest


def test_config():
    if os.path.exists('T_release/'):
        shutil.rmtree('T_release/')
    if os.path.exists('xHI_release/'):
        shutil.rmtree('xHI_release/')

    download().model()
    download(xHI=True).model()

    paramnames = [r'$\log(f_*)$', r'$\log(V_c)$', r'$\log(f_X)$',
                  r'$\nu_\mathrm{min}$', r'$\tau$', r'$\alpha$',
                  r'$R_\mathrm{mfp}$']

    # Providing this with global signal data as neutral fraction data is
    # not publicly available. Will not effect efficacy of the test.
    config('xHI_release/', paramnames, '21cmGEM_data/',
           logs=[0, 1, 2])

    assert(os.path.exists('xHI_release/gui_configuration.csv') is True)

    res = np.genfromtxt(
            'xHI_release/gui_configuration.csv',
            delimiter=',',
            names=True,
            dtype='U100,f8,f8,f8,f8,U100,U100',  # Explicitly specify types
            encoding='utf-8'
        )

    logs = res['logs'].tolist()
    logs = [int(x) for x in logs if x != '--']
    assert(logs == [0, 1, 2])

    paramnames = [r'$\log(f_*)$', r'$\log(V_c)$', r'$\log(f_X)$',
                  r'$\tau$', r'$\alpha$', r'$\nu_\mathrm{min}$',
                  r'$R_\mathrm{mfp}$']

    config('T_release/', paramnames, '21cmGEM_data/')

    assert(os.path.exists('T_release/gui_configuration.csv') is True)

    with pytest.raises(KeyError):
        config('T_release', paramnames, '21cmGEM_data/')
    with pytest.raises(TypeError):
        config(10, paramnames, '21cmGEM_data/')
    with pytest.raises(KeyError):
        config('T_release/', paramnames, '21cmGEM_data')
    with pytest.raises(TypeError):
        config('T_release/', paramnames, '21cmGEM_data/', logs='banana')
    with pytest.raises(TypeError):
        config('T_release/', 5, '21cmGEM_data/')

    with pytest.raises(KeyError):
        config('T_release/', paramnames, '21cmGEM_data/', color='C0')



import numpy as np
from globalemu.preprocess import process
import os
import pytest
import shutil


def test_preprocess_integer_redshift():
    z = np.arange(5, 5 + 451, dtype=np.uint16)
    process(10, z, data_location='21cmGEM_data/')
    assert not np.any(np.isnan(np.loadtxt('model_dir/AFB.txt')))


def test_preprocess():
    z = np.arange(5, 50.1, 0.1)

    process(10, z, data_location='21cmGEM_data/')
    process(10, z, data_location='21cmGEM_data/', xHI=True)
    process('full', z, data_location='21cmGEM_data/')
    process('full', z, data_location='21cmGEM_data/', AFB=False)
    process(10, z, data_location='21cmGEM_data/', resampling=False)

    files = ['AFB_norm_factor.npy', 'AFB.txt', 'cdf.txt', 'data_maxs.txt',
             'data_mins.txt', 'indices.txt', 'labels_stds.npy', 'samples.txt',
             'train_data.txt', 'train_dataset.csv', 'train_label.txt', 'z.txt',
             'preprocess_settings.pkl']

    for i in range(len(files)):
        assert(os.path.exists('model_dir/' + files[i]) is True)

    full_train_data = np.loadtxt(
        'model_dir/train_dataset.csv', dtype=float,
        delimiter=',')

    for i in range(full_train_data.shape[1]):
        if i < full_train_data.shape[1] - 1:
            assert(np.isclose(full_train_data[:, i].min(),
                   0, rtol=1e-1, atol=1e-1))
            assert(np.isclose(full_train_data[:, i].max(),
                   1, rtol=1e-1, atol=1e-1))

    with pytest.raises(TypeError):
        process(10.2, z, data_location='21cmGEM_data/')
    with pytest.raises(TypeError):
        process(10, 10, data_location='21cmGEM_data/')
    with pytest.raises(KeyError):
        process(10, z, data_location='data_download')
    with pytest.raises(TypeError):
        process(10, z, data_location='21cmGEM_data/', base_dir=10)
    with pytest.raises(TypeError):
        process(10, z, data_location='21cmGEM_data/', logs=True)
    with pytest.raises(TypeError):
        process(10, z, data_location='21cmGEM_data/', AFB=10)
    with pytest.raises(TypeError):
        process(10, z, data_location='21cmGEM_data/', resampling=10)
    with pytest.raises(TypeError):
        process(10, z, data_location='21cmGEM_data/', resampling=10)

    dir = ['21cmGEM_data/', 'model_dir/']
    for i in range(len(dir)):
        if os.path.exists(dir[i]):
            shutil.rmtree(dir[i])



from globalemu.downloads import download
import os
import pytest
import requests
import numpy as np


def download_21cmGEM_data():
    data_dir = '21cmGEM_data/'
    if not os.path.exists(data_dir):
        os.mkdir(data_dir)

    files = ['Par_test_21cmGEM.txt',
             'Par_train_21cmGEM.txt',
             'T21_test_21cmGEM.txt',
             'T21_train_21cmGEM.txt']
    saves = ['test_data.txt',
             'train_data.txt',
             'test_labels.txt',
             'train_labels.txt']

    for i in range(len(files)):
        url = 'https://zenodo.org/record/4541500/files/' + files[i]
        with open(data_dir + saves[i], 'wb') as f:
            f.write(requests.get(url).content)

    td = np.loadtxt(data_dir + 'train_data.txt', dtype=float) 
    tl = np.loadtxt(data_dir + 'train_labels.txt', dtype=float)

    np.savetxt(data_dir + 'train_data.txt', td[:500, :])
    np.savetxt(data_dir + 'train_labels.txt', tl[:500, :])


def test_existing_dir():
    if os.path.exists('kappa_HH.txt'):
        os.remove('kappa_HH.txt')

    download().kappa()

    with pytest.raises(TypeError):
        download(xHI=2).kappa()

    # for use in later tests...
    download_21cmGEM_data()



"""
This function can be used to generate a configurate file for the GUI that
is specific to a given trained model. The file gets saved into the supplied
``base_dir`` which should contain the relevant trained model. The user
also needs to supply a path to the ``data_dir`` that contains the relevant
testing and training data. Additional arguments are described below.

A GUI config file is required to be able to visualise the signals with the
GUI and once generated the gui can be run from the command line

.. code:: bash

    globalemu /path/to/base_dir/containing/model/and/config/

"""

import numpy as np
import pickle


class config():

    r"""

    **Parameters:**

        base_dir: **string**
            | The path to the file containing the trained tensorflow model
                that the user wishes to visualise with the GUI. Must end
                in '/'.

        paramnames: **list of strings**
            | This should be a list of parameter names in the correct input
                order. For example for the released global signal model this
                would correspond to

                .. code: python

                    paramnames = [r'$\log(f_*)$', r'$\log(V_c)$',
                                  r'$\log(f_X)$',
                                  r'$\tau$', r'$\alpha$',
                                  r'$\nu_\mathrm{min}$',
                                  r'$R_\mathrm{mfp}$']

                Latex strings can be provided as above.

        data_dir: **string**
            | The file path to the training and test data which is used to set
                the y lims of the GUI graph and ranges/intervals of GUI
                sliders.

    **Kwargs:**

        logs: **list / default: [0, 1, 2]**
            | The indices corresponding to the astrophysical
                parameters that
                were logged during training. The default assumes
                that the first three columns in "train_data.txt" are
                :math:`{f_*}` (star formation efficiency),
                :math:`{V_c}` (minimum virial circular velocity) and
                :math:`{f_x}` (X-ray efficieny).

        ylabel: **string / default: 'y'**
            | y-axis label for gui plot.

    """

    def __init__(self, base_dir, paramnames, data_dir, **kwargs):

        for key, values in kwargs.items():
            if key not in set(
                    ['logs', 'ylabel']):
                raise KeyError("Unexpected keyword argument in process()")

        self.base_dir = base_dir
        self.paramnames = paramnames
        self.data_dir = data_dir
        self.logs = kwargs.pop('logs', [0, 1, 2])
        self.ylabel = kwargs.pop('ylabel', 'y')

        file_kwargs = [self.base_dir, self.data_dir]
        file_strings = ['base_dir', 'data_dir']
        for i in range(len(file_kwargs)):
            if type(file_kwargs[i]) is not str:
                raise TypeError("'" + file_strings[i] + "' must be a sting.")
            elif file_kwargs[i].endswith('/') is False:
                raise KeyError("'" + file_strings[i] + "' must end with '/'.")

        file = open(self.base_dir + "preprocess_settings.pkl", "rb")
        self.preprocess_settings = pickle.load(file)

        if type(self.paramnames) is not list:
            raise TypeError("'paramnames' must be a list of strings.")

        if type(self.logs) is not list:
            raise TypeError("'logs' must be a list.")

        test_data = np.loadtxt(data_dir + 'test_data.txt')
        test_labels = np.loadtxt(data_dir + 'test_labels.txt')
        for i in range(test_data.shape[1]):
            if i in self.logs:
                for j in range(test_data.shape[0]):
                    if test_data[j, i] == 0:
                        test_data[j, i] = 1e-6
                test_data[:, i] = np.log10(test_data[:, i])

        data_mins = test_data.min(axis=0)
        data_maxs = test_data.max(axis=0)

        full_logs = []
        for i in range(len(data_maxs)):
            if i in set(self.logs):
                full_logs.append(i)
            else:
                full_logs.append('--')

        # Create the data arrays
        n = len(data_maxs)
        names = np.array(self.paramnames, dtype='U100')  # Unicode strings
        mins = np.array(data_mins, dtype=float)
        maxs = np.array(data_maxs, dtype=float)

        # label_min: first value is test_labels.min(), rest are empty strings
        label_min = np.array([test_labels.min()] + ['']*(n-1), dtype='U100')

        # label_max: first value is test_labels.max(), rest are empty strings
        label_max = np.array([test_labels.max()] + ['']*(n-1), dtype='U100')

        logs = np.array(full_logs, dtype='U100')
        ylabel = np.array([self.ylabel]*n, dtype='U100')

        # Stack into 2D array (transpose to get columns)
        data = np.column_stack([names, mins, maxs, label_min, label_max, logs, ylabel])

        # Write header
        header = 'names,mins,maxs,label_min,label_max,logs,ylabel'

        # Save to CSV
        np.savetxt(
            base_dir + 'gui_configuration.csv',
            data,
            delimiter=',',
            header=header,
            comments='',  # Prevents '#' being added to header
            fmt='%s'      # String format for all columns
        )



import numpy as np
from globalemu.eval import evaluate
from globalemu.plotter import signal_plot
from globalemu.downloads import download
import os
import shutil
import pytest

params = [0.25, 30, 2, 0.056, 1.3, 2, 30]
z = np.arange(10, 20, 100)


def test_existing_dir():
    if os.path.exists('T_release/'):
        shutil.rmtree('T_release/')
    if os.path.exists('xHI_release/'):
        shutil.rmtree('xHI_release/')

    download().model()
    download(xHI=True).model()

    predictor = evaluate(base_dir='T_release/')

    parameters = np.loadtxt('21cmGEM_data/test_data.txt')
    labels = np.loadtxt('21cmGEM_data/test_labels.txt')

    signal_plot(parameters, labels, 'rmse',
                predictor, 'T_release/', figsizex=3.15,
                figsizey=6, loss_label='RMSE = {:.4f} [mK]')

    assert(os.path.exists('T_release/eval_plot.pdf') is True)

    with pytest.raises(KeyError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release/', figsize_x=3.15)

    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release')
    with pytest.raises(TypeError):
        signal_plot(1, labels, 'rmse',
                    predictor, 'T_release/')
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'banana',
                    predictor, 'T_release/')
    with pytest.raises(TypeError):
        signal_plot(parameters, 6, 'rmse',
                    predictor, 'T_release/')
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    'foobar', 'T_release/')
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release/', rtol='banana')
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release/', atol='banana')
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release/', figsizex='banana')
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release/', figsizey='banana')
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release/', loss_label=60)
    with pytest.raises(TypeError):
        signal_plot(parameters, labels, 'rmse',
                    predictor, 'T_release/', xHI=60)

    def loss_func(labels, signals):
        return np.mean(np.square(labels - signals))*1000

    signal_plot(parameters, labels, loss_func,
                predictor, 'T_release/')

    assert(os.path.exists('T_release/eval_plot.pdf') is True)

    signal_plot(parameters, labels, 'mse',
                predictor, 'T_release/')

    signal_plot(parameters, labels, 'GEMLoss',
                predictor, 'T_release/')



import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button
import numpy as np
import sys
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
from globalemu.eval import evaluate

def main():
    base_dir = sys.argv[1]
    if base_dir[-1] != '/':
        base_dir += '/'
    
    config = np.genfromtxt(
        base_dir + 'gui_configuration.csv',
        delimiter=',',
        names=True,
        dtype='U100,f8,f8,f8,f8,U100,U100',  # Explicitly specify types
        encoding='utf-8'
    )
    
    logs = config['logs'].tolist()
    logs = [int(x) for x in logs if x != '--']
    label_min = config['label_min'][0]
    label_max = config['label_max'][0]
    ylabel = config['ylabel'][0]
    
    predictor = evaluate(base_dir=base_dir, logs=logs)
    
    # Calculate initial values (center of range)
    center = []
    for i in range(len(config['names'])):
        center.append(config['mins'][i] + (config['maxs'][i] - config['mins'][i])/2)
    center = np.array(center)
    
    # Create figure with plot on left, sliders on right
    n_sliders = len(config['names'])
    fig = plt.figure(figsize=(14, 8))
    
    # Main plot on left side - takes full height
    ax_plot = plt.subplot2grid((n_sliders, 2), (0, 0), rowspan=n_sliders)
    
    # Initial signal
    def get_params(slider_vals):
        params = []
        for i in range(len(slider_vals)):
            if i in set(logs):
                params.append(10**slider_vals[i])
            else:
                params.append(slider_vals[i])
        return params
    
    signal, z = predictor(get_params(center))
    line, = ax_plot.plot(z, signal, c='k', lw=2)
    ax_plot.set_xlabel('z')
    ax_plot.set_ylabel(ylabel)
    ax_plot.set_ylim([label_min, label_max])
    ax_plot.grid(True, alpha=0.3)
    
    # Create sliders on right side
    sliders = []
    
    for i in range(n_sliders):
        # Create axis for this slider on right column
        ax_slider = plt.subplot2grid((n_sliders, 2), (i, 1))
        
        # Create slider
        slider = Slider(
            ax_slider,
            config['names'][i],
            config['mins'][i],
            config['maxs'][i],
            valinit=center[i],
            valstep=(config['maxs'][i] - config['mins'][i])/100
        )
        sliders.append(slider)
    
    # Update function
    def update(val):
        slider_vals = [s.val for s in sliders]
        params = get_params(slider_vals)
        signal, z = predictor(params)
        line.set_ydata(signal)
        fig.canvas.draw_idle()
    
    # Connect all sliders to update function
    for slider in sliders:
        slider.on_changed(update)
    
    # Reset button
    ax_reset = plt.axes([0.8, 0.025, 0.1, 0.04])
    btn_reset = Button(ax_reset, 'Reset')
    
    def reset(event):
        for i, slider in enumerate(sliders):
            slider.set_val(center[i])
    
    btn_reset.on_clicked(reset)
    
    plt.tight_layout()
    plt.show()

if __name__ == '__main__':
    main()


import numpy as np
from globalemu.eval import evaluate
from globalemu.downloads import download
import os
import shutil
import pytest

params = [0.25, 30, 2, 0.056, 1.3, 2, 30]
originalz = np.linspace(10, 20, 100)


def test_existing_dir():
    if os.path.exists('T_release/'):
        shutil.rmtree('T_release/')
    if os.path.exists('xHI_release/'):
        shutil.rmtree('xHI_release/')

    download().model()
    download(xHI=True).model()

    predictor = evaluate(z=originalz,
                         base_dir='T_release/', gc=True)
    signal, z = predictor(params)

    assert(len(signal.shape) == 1)

    params_array = []
    for i in range(10):
        params_array.append(np.random.uniform(0, 1, 7))
    params_array = np.array(params_array)

    data_max = np.loadtxt('T_release/data_maxs.txt')
    data_min = np.loadtxt('T_release/data_mins.txt')

    for i in range(params_array.shape[1]):
        params_array[:, i] = params_array[:, i] * \
            (data_max[i] - data_min[i]) + data_min[i]

    signal, z = predictor(params_array)

    assert(len(signal.shape) == 2)

    predictor = evaluate(z=10, base_dir='T_release/')
    signal, z = predictor(params)
    preprocess = predictor.preprocess_settings
    assert(preprocess['AFB'] is True)
    assert(preprocess['resampling'] is True)
    assert(preprocess['std_division'] is True)

    predictor = evaluate(z=originalz, base_dir='xHI_release/')
    signal, z = predictor(params)

    with pytest.raises(KeyError):
        evaluate(z=originalz, basedir='T_release/')

    with pytest.raises(TypeError):
        predictor = evaluate(z=originalz, base_dir='T_release/')
        predictor(10)
    with pytest.raises(TypeError):
        predictor = evaluate(z=originalz, base_dir=100)
    with pytest.raises(TypeError):
        predictor = evaluate(z=originalz, base_dir='T_release/', logs=10)
    with pytest.raises(TypeError):
        predictor = evaluate(z='foo', base_dir='T_release/')
    with pytest.raises(TypeError):
        predictor = evaluate(z=originalz, base_dir='T_release/',
                             gc='false')



from globalemu.cmSim import calc_signal
import numpy as np
import os
import shutil


def test_calc_signal():
    if os.path.exists('kappa_HH.txt'):
        os.remove('kappa_HH.txt')

    if not os.path.exists('model_dir/'):
        os.mkdir('model_dir/')

    os.system("cp T_release/AFB_norm_factor.npy model_dir/")

    _ = calc_signal(np.arange(5, 50, 100), 'model_dir/')

    shutil.rmtree('model_dir/')



import numpy as np


class sampling():

    r"""

    This function is used to resample the input training signals so that
    the regions of high variance in the training data are emphasised during
    training. The function finds the regions of high variance and
    builds a probability distribution from which to produce new redshift
    samples. See MNRAS preprint for details at
    https://arxiv.org/abs/2104.04336.

    **Parameters:**

    z: **list or np.array**
        | The original redshift data points.

    base_dir: **string**
        | The location of the trained model files. The samples and cdf are
            needed during evaluation of the network and consequently they
            are saved when this class is called.

    train_labels: **array**
        | The training global 21-cm or neutral fraction signals to resample
            at the new redshifts.

    """
    def __init__(self, z, base_dir, train_labels):
        self.z = z
        self.base_dir = base_dir
        self.train_labels = train_labels

        difference = []
        for i in range(self.train_labels.shape[1]):
            difference.append(
                self.train_labels[:, i].max() - self.train_labels[:, i].min())
        difference = np.array(difference)

        difference = [
            difference[i]/np.sum(difference) for i in range(len(difference))]

        self.cdf = np.cumsum(difference)

        x = np.linspace(0, 1, len(self.z))

        self.samples = np.interp(x, self.cdf, self.z)
        np.savetxt(self.base_dir + 'samples.txt', self.samples)
        np.savetxt(self.base_dir + 'cdf.txt', self.cdf)



import numpy as np
from globalemu.preprocess import process
from globalemu.network import nn
from tensorflow.keras import backend as K
import os
import shutil
import pytest


def test_process_nn():

    def custom_loss(y, y_, x):
        return K.mean(K.abs(y - y_))

    z = np.arange(5, 50.1, 0.1)

    process(10, z, data_location="21cmGEM_data/")
    nn(batch_size=451, layer_sizes=[8], epochs=10, loss_function=custom_loss)

    # results of below will not make sense as it is being run on the
    # global signal data but it will test the code (xHI data not public)
    process(10, z, data_location="21cmGEM_data/", xHI=True)
    nn(batch_size=451, layer_sizes=[8], epochs=5, xHI=True)

    nn(batch_size=451, layer_sizes=[8], epochs=5, output_activation="linear")

    # test early_stop code
    nn(batch_size=451, layer_sizes=[], epochs=20, early_stop=True)

    process(10, z, data_location="21cmGEM_data/", base_dir="base_dir/")
    nn(
        batch_size=451,
        layer_sizes=[],
        random_seed=10,
        epochs=30,
        base_dir="base_dir/",
        early_stop=True,
    )

    dir = ["model_dir/", "base_dir/"]
    for i in range(len(dir)):
        if os.path.exists(dir[i]):
            shutil.rmtree(dir[i])


def test_process_nn_keyword_errors():
    z = np.arange(5, 50.1, 0.1)
    with pytest.raises(KeyError):
        process(10, z, datalocation="21cmGEM_data/")

    with pytest.raises(KeyError):
        nn(batch_size=451, layersizes=[8], epochs=10)


@pytest.mark.parametrize(
    "keyword, value",
    [
        ("batch_size", "foo"),
        ("activation", 10),
        ("epochs", False),
        ("lr", "bar"),
        ("dropout", True),
        ("input_shape", "foo"),
        ("output_shape", "foobar"),
        ("layer_sizes", 10),
        ("base_dir", 50),
        ("early_stop", "foo"),
        ("xHI", "false"),
        ("resume", 10),
        ("output_activation", 2),
        ("loss_function", "foobar"),
    ],
)
def test_process_nn_type_errors(keyword, value):
    z = np.arange(5, 50.1, 0.1)

    process(10, z, data_location="21cmGEM_data/")
    with pytest.raises((TypeError, ValueError)):
        nn(**{keyword: value})
